#include "nmEnvelope.hpp"

void nmEnvelope::noteOn(){
  level = 0.0f;
  state = ATTACK_TIME;

  if(attack_time == 0.0f){
    state = DECAY_TIME;
    if(decay_time == 0.0f){
      state = SUSTAIN_TIME;
    }
    else{
      level = attack_level;
    }
  }
  if(attack_time == 0.0f){
    d_at = 0.0f;
  }
  else{
    d_at = attack_level / attack_time;
  }
  if(decay_time == 0.0f){
    d_dt = 0.0f;
  }
  else{
    d_dt = (attack_level - sustain_level) / decay_time;
  }

  at_count = 0.0f;
  dt_count = 0.0f;
}

void nmEnvelope::noteOff(){
  state = RELEASE_TIME;

  if(release_time == 0.0f){
    d_rt = 0.0f;
  }
  else{
    d_rt = level / release_time;
  }

  rt_count = 0.0f;
}

float nmEnvelope::getEnvelopeLevel(){
  switch(state){
    case SILENCE:
      level = 0.0f;
      break;
    case ATTACK_TIME:
      level += d_at;
      at_count += 1.0f;
      if(attack_time - at_count < 0){
        state = DECAY_TIME;
      }
      break;
    case DECAY_TIME:
      level -= d_dt;
      dt_count += 1.0f;
      if(decay_time - dt_count < 0){
        state = SUSTAIN_TIME;
      }
      break;
    case SUSTAIN_TIME:
      level = sustain_level;
      break;
    case RELEASE_TIME:
      level -= d_rt;
      rt_count += 1.0f;
      if(release_time - rt_count < 0){
        state = SILENCE;
        level = 0.0f;
      }
      break;
    default:
      level = 0.0f;
      break;
  }

  return level;
}
