#ifndef __NMSYNTHCONTROL__
#define __NMSYNTHCONTROL__

#include "nmSynthNote.hpp"

class nmSynthControl
{
  enum{max_note = 8};
  nmSynthNote note[max_note];
public:
  nmSynthControl(){};
  ~nmSynthControl(){};
  void setSampleRate(float sr){
    for(int i = 0; i < max_note; i++){
      note[i].setSampleRate(sr);
    }
  }
  void setWaveType(const int &wt){
    for(int i = 0; i < max_note; i++){
      note[i].setWaveType(wt);
    }
  }
  void setLowPass(const float &lp, const float &res){
    for(int i = 0; i < max_note; i++){
      note[i].setSynthNoteLowPass(lp, res);
    }
  }
  void setEnvelope(const float &at, const float &al, const float &dt, const float &sl, const float &rt){
    for(int i = 0; i < max_note; i++){
      note[i].setSynthNoteEnvelope(at, al, dt, sl, rt);
    }
  }
  void setPitch(const int &p){
    for(int i = 0; i < max_note; i++){
      note[i].setSynthNotePitch(((float)p - 64.0f) * 0.0156f);
    }
  }
  void setVibrato(const int &v){
    for(int i = 0; i < max_note; i++){
      note[i].setSynthNoteVibrato((float)v * 0.0002f);
    }
  }
  bool isSilence(){
    for(int i = 0; i < max_note; i++){
      if(note[i].isSilence() != true){
        return false;
      }
    }
    return true;
  }
  void setNote(const int &n, const int &velocity){
    for(int i = 0; i < max_note; i++){
      if(note[i].isSilence() == true){
        note[i].noteOn(n, (float)velocity * (1.0f / 127.0f));
        break;
      }
    }
  }
  void offNote(int n){
    for(int i = 0; i < max_note; i++){
      if(note[i].getNoteNum() - n == 0 && note[i].isNoteOn() == true){
        note[i].noteOff();
        break;
      }
    }
  }
  void offAllNote(){
    for(int i = 0; i < max_note; i++){
      note[i].noteOff();
    }
  }
  float getWave(){
    float w = 0.0f;

    for(int i = 0; i < max_note; i++){
      w += note[i].getWave();
    }
    return w;
  }
};

#endif // __NMSYNTHCONTROL__