#include <windows.h>
#include <commctrl.h>
#include <stdio.h>
#include <GL/gl.h>
#include <math.h>

#include "AudioEffectx.h"
#include "nmSynthEdit.hpp"
#include "nmSynth.hpp"
#include "dial.h"

extern HINSTANCE hInstance;

#define WIDTH 300
#define HEIGHT 200

static const int dial_w = 20;
static const int dial_h = 20;

static const int max_edit = 8;

typedef struct{
  HWND hwnd;
  nmSynth *nms;
}synth;

synth synth_edit[max_edit];

static int xPos, yPos;

static int dial_point[8][2] = {
  {67, 85},
  {177, 85},
  {233, 85},
  {32, 135},
  {93, 135},
  {149, 135},
  {205, 135},
  {264, 135}
};

/* sNZtH[}bg̐ݒ */
static BOOL bSetupPixelFormat(HDC hdc);

HWND ghwnd;

GLuint texName[1];

GLubyte *image0;

/* eNX`ݒ */
void initTexture(void){
  GLubyte *p;

  image0 = (GLubyte *)malloc(3 * 512 * 512 * sizeof(unsigned char));

  if(image0 == NULL){
    return;
  }

  HRSRC hRsrc =  FindResource(hInstance, "IDR_BIN1", "BIN");
  HANDLE hRes = LoadResource(hInstance, hRsrc);

  if(hRes != NULL){
    p = (GLubyte *)LockResource(hRes);

    for(int i = 0; i < 512 * 512; i++){
      image0[i * 3] = p[i * 3];
      image0[i * 3 + 1] = p[i * 3 + 1];
      image0[i * 3 + 2] = p[i * 3 + 2];
    }
  }
}

void deinitTexture(void)
{
  if(image0 != NULL){
    free(image0);
  }
}

void dial(void)
{
  int i, j;

  /* WWp */
  GLdouble cube_vertex[][3] = {
    { -0.1, 0.8, 0.6 },
    { -0.1, 0.0, 0.6 },
    { 0.1, 0.0, 0.6 },
    { 0.1, 0.8, 0.6 },
    { -0.1, 0.9, 0.0 },
    { -0.1, 0.0, 0.0 },
    { 0.1, 0.0, 0.0 },
    { 0.1, 0.9, 0.0 },
  };

  /* WȂŖʂf[^Up */
  int cube_face[][4] = {
    { 0, 1, 2, 3 },
    { 4, 0, 3, 7 },
    { 7, 3, 2, 6 },
    { 6, 2, 1, 5 },
    { 5, 1, 0, 4 },
    { 7, 6, 5, 4 }
  };

  /* ʂƂ̖@f[^ */
  GLdouble cube_normal[][3] = {
    { 0.0, 0.0, 1.0 },
    { 0.0, 1.0, 0.0 },
    { 1.0, 0.0, 0.0 },
    { 0.0, -1.0, 0.0 },
    { -1.0, 0.0, 0.0 },
    { 0.0, 0.0, -1.0 }
  };

  /* ʂƂ̐Ff[^ */
  GLfloat color[][4] = {
    { 1.0, 0.0, 0.0, 1.0 },
  };

  glBegin(GL_QUADS);

  for (j = 0; j < 64; j++) {
    glMaterialfv(GL_FRONT, GL_DIFFUSE, color[0]);

    glNormal3dv(dial_normal[j]);

    for (i = 0; i < 4; i++) {
      glVertex3dv(dial_vertex[dial_face[j][i]]);
    }
  }

  glEnd();

  glBegin(GL_QUADS);

  for (j = 0; j < 6; j++) {
    glMaterialfv(GL_FRONT, GL_DIFFUSE, color[0]);
    glNormal3dv(cube_normal[j]);
    for (i = 0; i < 4; i++) {
      glVertex3dv(cube_vertex[cube_face[j][i]]);
    }
  }

  glEnd();
}

void paint_dial(double x, double y, double z, double r){
  glPushMatrix();

  glTranslated(x, y, z);

  glRotated(r, 0.0, 0.0, 1.0);

  dial();

  glPopMatrix();
}

/* ̏ */
void nmgl_init(void){
  GLfloat lightAmb[] = {0.0, 0.0, 0.0, 1.0};
  GLfloat lightPos[] = {0.0, 0.0, -1.0, 1.0};
  GLfloat lightDif[] = {1.0, 1.0, 1.0, 1.0};

  /* wiF̎w RGBA */
  glClearColor(0.0, 0.0, 0.0, 0.0);

  /* fvXobt@̗L */
  glEnable(GL_DEPTH_TEST);

  /* JO̗L */
  glEnable(GL_CULL_FACE);
  /* Culling GL_BACK or GL_FRONT */
  glCullFace(GL_BACK);

  /* ̐ݒ */
  glLightfv(GL_LIGHT0, GL_AMBIENT, lightAmb);
  glLightfv(GL_LIGHT0, GL_POSITION, lightPos);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, lightDif);

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);

  glGenTextures(1, texName);
  glBindTexture(GL_TEXTURE_2D, texName[0]);

  /* eNX`̐ݒ */
  glTexImage2D(
    GL_TEXTURE_2D
    ,0                 /* ~bv}bv */
    ,3                 /* RGBȂ̂3 */
    ,512               /*  */
    ,512               /*  */
    ,0                 /* E̕ */
    ,GL_RGB            /* RGB` */
    ,GL_UNSIGNED_BYTE  /* oCgf[^ */
    ,image0);          /* C[Wւ̃|C^ */

  /* SɌJԂGL_REPEAT ł؂ GL_CLAMP */
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  /* TɌJԂGL_REPEAT ł؂ GL_CLAMP */
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

  /* 摜kɉ摜͋߂FI */
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  /* 摜g厞ɉ摜͋߂FI */
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

  /* ̖ʂ̐Fւ̑Ή f GL_MODULATE  GL_DECAL */
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
}

void nmgl_deinit(void){
  /* eNX`̉ */
  glDeleteTextures(1, texName);
}

/* TCY̏ */
void nmgl_resize(int w, int h){
  /* r[|[g̐ݒ */
  glViewport(0, 0, w, h);

  /* ˉeϊ[h */
  glMatrixMode(GL_PROJECTION);

  /* ϊsPʍs */
  glLoadIdentity();

  /* ˉe̎` AEAAAOA */
  glFrustum(-5.32, 5.32, -4.0, 4.0, 19.0, 20.0);

  /* fϊ[hɖ߂ */
  glMatrixMode(GL_MODELVIEW);
}

/* `̏ */
void nmgl_paint(nmSynthEdit *nmse){

  // wiFŃNA
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glLoadIdentity();

  paint_dial(-3.0, 0.6f, -20.0, nmse->getDialWT());
  paint_dial(1.0, 0.6f, -20.0, nmse->getDialLP());
  paint_dial(3.0, 0.6f, -20.0, nmse->getDialRes());
  paint_dial(-4.0, -1.4f, -20.0, nmse->getDialAT());
  paint_dial(-2.0, -1.4f, -20.0, nmse->getDialAL());
  paint_dial(0.0, -1.4f, -20.0, nmse->getDialDT());
  paint_dial(2.0, -1.4f, -20.0, nmse->getDialSL());
  paint_dial(4.0, -1.4f, -20.0, nmse->getDialRT());

  glPushMatrix();

  glEnable(GL_TEXTURE_2D);

  glBindTexture(GL_TEXTURE_2D , texName[0]);

  glTranslated(0.0, 0.0, -19.9);

  GLfloat color[][4] = {
    { 1.0, 1.0, 1.0, 1.0 },
  };

  glBegin(GL_QUADS);

  // wipl
  glBegin(GL_QUADS);

  glMaterialfv(GL_FRONT, GL_DIFFUSE, color[0]);

  glNormal3d(0.0, 0.0, 1.0);
  glTexCoord2f(0.0, 0.0);
  glVertex3d(-5.6, 4.21, 0.0);
  glTexCoord2f(0.0, 1.0);
  glVertex3d(-5.6, -4.21, 0.0);
  glTexCoord2f(1.0, 1.0);
  glVertex3d(5.6, -4.21, 0.0);
  glTexCoord2f(1.0, 0.0);
  glVertex3d(5.6, 4.21, 0.0);

  glEnd();

  glDisable(GL_TEXTURE_2D);

  glPopMatrix();

  glFinish();
}


typedef struct{
  HWND hwnd;
  HGLRC hrc;
}window;

/* vV[W */
LONG WINAPI nmSynthEdit::WindowProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
  HINSTANCE hInst;
  hInst = (HINSTANCE)GetWindowLong(hwnd, GWL_HINSTANCE);

  //nmSynthEdit *nmse = (nmSynthEdit *)GetWindowLong(hwnd, GWL_USERDATA);

  static window synth_window[max_edit];

  static int active_dial;
  static int cxPos, cyPos;

  static BOOL bCap = FALSE;

  static int edit_num = 0;

  int current_edit;
  int current_window;

  for(current_edit = 0; current_edit < max_edit; current_edit++){
    if(hwnd == synth_edit[current_edit].hwnd){
      break;
    }
  }

  for(current_window = 0; current_window < max_edit; current_window++){
    if(hwnd == synth_window[current_window].hwnd){
      break;
    }
  } 

  HDC chdc = GetDC(hwnd);

  wglMakeCurrent(chdc, synth_window[current_window].hrc);

  switch (message){
    case WM_CREATE:
      {
        int i;

        for(i = 0; i < max_edit; i++){
          if(synth_window[i].hrc == NULL){
            HDC hdc = GetDC(hwnd);
            bSetupPixelFormat(hdc);
            synth_window[i].hrc = wglCreateContext(hdc);
            wglMakeCurrent(hdc, synth_window[i].hrc);
            synth_window[i].hwnd = hwnd;
            // 
            nmgl_init();
            break;
          }
        }
      }
      break;
    case WM_LBUTTONDOWN:
      {
        POINTS pts = MAKEPOINTS(lParam);

        cxPos = pts.x;
        cyPos = pts.y;

        for(int i = 0; i <= 8; i++){
          if(i == 8){
            active_dial = 8;
            break;
          }

          if(dial_point[i][0] - dial_w < cxPos && dial_point[i][0] + dial_w > cxPos && dial_point[i][1] - dial_h < cyPos && dial_point[i][1] + dial_h > cyPos){
            active_dial = i;
            bCap = TRUE;
            SetCapture(hwnd);
            break;
          }
        }

        nmSynthEdit *nmse = synth_edit[current_edit].nms->getNMSE();

        switch(active_dial){
          case 0:
            cxPos += (int)nmse->getDialWT();
            cyPos += (int)nmse->getDialWT();
            break;
          case 1:
            cxPos += (int)nmse->getDialLP();
            cyPos += (int)nmse->getDialLP();
            break;
          case 2:
            cxPos += (int)nmse->getDialRes();
            cyPos += (int)nmse->getDialRes();
            break;
          case 3:
            cxPos += (int)nmse->getDialAT();
            cyPos += (int)nmse->getDialAT();
            break;
          case 4:
            cxPos += (int)nmse->getDialAL();
            cyPos += (int)nmse->getDialAL();
            break;
          case 5:
            cxPos += (int)nmse->getDialDT();
            cyPos += (int)nmse->getDialDT();
            break;
          case 6:
            cxPos += (int)nmse->getDialSL();
            cyPos += (int)nmse->getDialSL();
            break;
          case 7:
            cxPos += (int)nmse->getDialRT();
            cyPos += (int)nmse->getDialRT();
            break;
          default:
            break;
        }
      }
      break;
    case WM_MOUSEMOVE:
      if(bCap){
        POINTS pts = MAKEPOINTS(lParam);
        SetCursor(LoadCursor(NULL, IDC_CROSS));

        xPos = cxPos - pts.x;
        yPos = cyPos - pts.y;

        int r = xPos;

        if(r < -120)r = -120;
        if(r > 120)r = 120;

        nmSynthEdit *nmse = synth_edit[current_edit].nms->getNMSE();

        switch(active_dial){
          case 0:
            nmse->setDialWT((double)r);
            break;
          case 1:
            nmse->setDialLP((double)r);
            break;
          case 2:
            nmse->setDialRes((double)r);
            break;
          case 3:
            nmse->setDialAT((double)r);
            break;
          case 4:
            nmse->setDialAL((double)r);
            break;
          case 5:
            nmse->setDialDT((double)r);
            break;
          case 6:
            nmse->setDialSL((double)r);
            break;
          case 7:
            nmse->setDialRT((double)r);
            break;
          default:
            break;
        }
        //ĕ`
        InvalidateRect(hwnd, NULL, FALSE);
      }
      else{
        SetCursor(LoadCursor(NULL, IDC_ARROW));
      }
      break;
    case WM_LBUTTONUP:
      {
        nmSynthEdit *nmse = synth_edit[current_edit].nms->getNMSE();

        synth_edit[current_edit].nms->setWaveType((float)(-nmse->getDialWT() + 120.0) * 0.00416f);
        synth_edit[current_edit].nms->setLowPass((float)(-nmse->getDialLP() + 120.0) * 0.00416f, (float)(-nmse->getDialRes() + 120.0) * 0.00416f);
        synth_edit[current_edit].nms->setEnvelope((float)(-nmse->getDialAT() + 120.0) * 0.00416f, (float)(-nmse->getDialAL() + 120.0) * 0.00416f, (float)(-nmse->getDialDT() + 120.0) * 0.00416f, (float)(-nmse->getDialSL() + 120.0) * 0.00416f, (float)(-nmse->getDialRT() + 120.0) * 0.00416f);

        SetCursor(LoadCursor(NULL, IDC_ARROW));
        ReleaseCapture();
        bCap = FALSE;
      }
      break;
    case WM_PAINT:
      {
        nmSynthEdit *nmse = synth_edit[current_edit].nms->getNMSE();

        /* ` */
        nmgl_paint(nmse);

        /* _uobt@] */
        SwapBuffers(chdc);
      }
      break;
    case WM_SIZE:
      {
        RECT rect;

        GetClientRect(hwnd, &rect);
        nmgl_resize(rect.right, rect.bottom);
      }
      break;
    case WM_DESTROY:
      {
        nmgl_deinit();
        // Еt
        wglMakeCurrent(chdc, 0);
        wglDeleteContext(synth_window[current_window].hrc);
        synth_window[current_window].hrc = NULL;
        ReleaseDC(hwnd, chdc);
      }
      break;
    default:
      break;
  }

  return DefWindowProc (hwnd, message, wParam, lParam);
}

/* RXgN^ */
nmSynthEdit::nmSynthEdit(nmSynth *effectx)
: AEffEditor(effectx), hwnd_e(NULL)
{
  int i;

  for(i = 0; i < max_edit; i++){
    if(synth_edit[i].nms == NULL){
      synth_edit[i].nms = effectx;
      break;
    }
  }

  if(i == max_edit){
    throw -1;
  }
}

/* fXgN^ */
nmSynthEdit::~nmSynthEdit(){
  int i;

  for(i = 0; i < max_edit; i++){
    if(synth_edit[i].nms != NULL){
      nmSynth *nms = synth_edit[i].nms;
      nmSynthEdit *nmse = nms->getNMSE();
      if(nmse == this){
        synth_edit[i].nms = NULL;
        break;
      }
    }
  }
}

void nmSynthEdit::setDial(const double &d0, const double &d1, const double &d2, const double &d3, const double &d4, const double &d5, const double &d6, const double &d7){
  dial_wave_type = -(d0 - 0.5) * 240.0;
  dial_attack_time = -(d3 - 0.5) * 240.0;
  dial_attack_level = -(d4 - 0.5) * 240.0;
  dial_decay_time = -(d5 - 0.5) * 240.0;
  dial_sustain_level = -(d6 - 0.5) * 240.0;
  dial_release_time = -(d7 - 0.5) * 240.0;
  dial_low_pass = -(d1 - 0.5) * 240.0;
  dial_resonance = -(d2 - 0.5) * 240.0;

  if(hwnd_e != NULL){
    InvalidateRect(hwnd_e, NULL, FALSE);
  }
}

long nmSynthEdit::getRect(ERect **erect){
  static ERect r = {0, 0, HEIGHT, WIDTH};
  *erect = &r;
  return true;
}

static int regist_count = 0;

long nmSynthEdit::open(void *ptr){
  systemWindow = ptr;

  if(regist_count == 0){
    WNDCLASS wd;
    wd.style = 0;
    wd.lpfnWndProc = WindowProc;
    wd.cbClsExtra = 0;
    wd.cbWndExtra = 0;
    wd.hInstance = hInstance;
    wd.hIcon = NULL;
    wd.hCursor = NULL;
    wd.hbrBackground = GetSysColorBrush(COLOR_BTNFACE);
    wd.lpszMenuName = NULL;
    wd.lpszClassName = "nmsynthClass";

    RegisterClass(&wd);

    initTexture();
  }

  regist_count++;

  HWND hwnd = CreateWindowEx(
                NULL,
                "nmsynthClass",
                "",
                WS_CHILD | WS_VISIBLE, 
                CW_USEDEFAULT,
                CW_USEDEFAULT,
                WIDTH,
                HEIGHT,
                (HWND)systemWindow,
                NULL,
                hInstance,
                NULL);

  hwnd_e = hwnd;

  int i;

  for(i = 0; i < max_edit; i++){
    if(synth_edit[i].nms != NULL){
      nmSynth *nms = synth_edit[i].nms;
      nmSynthEdit *nmse = nms->getNMSE();
      if(nmse == this){
        synth_edit[i].hwnd = hwnd;
        break;
      }
    }
  }

  SetWindowLong(hwnd, GWL_USERDATA, (long)this);

  return true;
}

void nmSynthEdit::close()
{
  hwnd_e = NULL;

  regist_count--;

  if(regist_count == 0){
    deinitTexture();

    UnregisterClass("nmsynthClass", hInstance);
  }
}

void nmSynthEdit::idle ()
{
  AEffEditor::idle ();
}

/* sNZtH[}bg̐ݒ */
BOOL bSetupPixelFormat(HDC hdc){
  static PIXELFORMATDESCRIPTOR pfd = {
    sizeof(PIXELFORMATDESCRIPTOR),  /* size of this pfd */
    1,                              /* version number */
    PFD_DRAW_TO_WINDOW |            /* support window */
    PFD_SUPPORT_OPENGL |            /* support OpenGL */
    PFD_DOUBLEBUFFER,               /* double buffered */
    PFD_TYPE_RGBA,                  /* RGBA type */
    24,                             /* 24-bit color depth */
    0, 0, 0, 0, 0, 0,               /* color bits ignored */
    0,                              /* no alpha buffer */
    0,                              /* shift bit ignored */
    0,                              /* no accumulation buffer */
    0, 0, 0, 0,                     /* accum bits ignored */
    32,                             /* 32-bit z-buffer */
    0,                              /* no stencil buffer */
    0,                              /* no auxiliary buffer */
    PFD_MAIN_PLANE,                 /* main layer */
    0,                              /* reserved */
    0, 0, 0                         /* layer masks ignored */
  };

  int pixelformat;

  if((pixelformat = ChoosePixelFormat(hdc, &pfd)) == 0){
    MessageBox(NULL, "ChoosePixelFormat failed", "Error", MB_OK);
    return FALSE;
  }

  if(SetPixelFormat(hdc, pixelformat, &pfd) == FALSE)    {
    MessageBox(NULL, "SetPixelFormat failed", "Error", MB_OK);
    return FALSE;
  }

  return TRUE;
}
