#ifndef __NMWAVE__
#define __NMWAVE__

class nmWave{
public:
  virtual ~nmWave(){};
  virtual void initWave(const float &freq, const float &samplerate) = 0;
  virtual float getWave(const float &index) = 0;
};

// `g
class nmRectWave : public nmWave{
  float point1, point2; // point1:ԓ_ point2:I_ 
public:
  nmRectWave(){}
  virtual ~nmRectWave(){};
  virtual void initWave(const float &freq, const float &samplerate){
    point2 = samplerate / freq;
    point1 = point2 * 0.5f;
  }
  virtual float getWave(const float &index){
    float p = index;
	
	float f = index / point2;

	p -= (int)f * point2;

    if(p - point1 < 0){
      return 1.0f;
    }
    else{
      return -1.0f;
    }
  }
};

// mRMg
class nmSawWave : public nmWave{
  float d;
  float point; // point:I_ 
public:
  nmSawWave(){}
  virtual ~nmSawWave(){};
  virtual void initWave(const float &freq, const float &samplerate){
    point = samplerate / freq;
    d = 2.0f / point;
  }
  virtual float getWave(const float &index){
    float p = index;
	
	float f = index / point;

	p -= (int)f * point;

    return p * d - 1.0f;
  }
};

// Opg
class nmTriWave : public nmWave{
  float d;
  float point1, point2, point3; // point1:1/4_ point2:3/4_ point3:I_ 
public:
  nmTriWave(){}
  virtual ~nmTriWave(){};
  virtual void initWave(const float &freq, const float &samplerate){
    point3 = samplerate / freq;
    point1 = point3 * 0.25f;
    point2 = point3 * 0.75f;
    d = 4.0f / point3;
  }
  virtual float getWave(const float &index){
    float p = index;
	
	float f = index / point3;

	p -= (int)f * point3;

    if(p - point1 < 0){
      return p * d;
    }
    else if(p - point2 < 0){
      return 2.0f - p * d;
    }
    else{
      return p * d - 3.0f;
    }
  }
};

#endif // __NMWAVE__
